@@ -7,7 +7,7 @@ import os
 import sys
 import glob
 import logging
-
+from itertools import chain
 log = logging.getLogger('byneuron')
 
 
@@ -109,6 +109,7 @@ class Byneuron:
         self._publicIdTypes = None
         self._isc = None
         # load indexSets for user and set one as default
+        self.indexsets = []
         self.indexsets = self.get_indexsets()
         self.indexset_active = self.indexsets[0]
 
@@ -166,6 +167,26 @@ class Byneuron:
         else:
             r.raise_for_status()
 
+    def get_indexset(self, x=None):
+        """
+        retrieves indexset entity
+        - active indexset
+        - name, key
+        - validates entity
+        :param x: input to check
+        :return: indexset or None
+        """
+        if x is None:
+            return self.indexset_active
+        if isinstance(x, str):
+            for i in self.indexsets:
+                if i.key == x or i.name == x:
+                    e = i
+                    return i
+        if isinstance(x, Entity) and x.entity_type == 'IndexSet':
+            return x
+
+
     def set_indexset(self, e, verbose=True):
         """
         analogy of selecting a tenant in frontend
@@ -173,12 +194,8 @@ class Byneuron:
         :param verbose: disable welcome message
         :return:
         """
-        if isinstance(e, str):
-            for i in self.indexsets:
-                if i.key == e or i.name == e:
-                    e = i
-                    break
-        if isinstance(e, Entity) and e.entity_type == 'IndexSet':
+        e = self.get_indexset(e)
+        if e is not None:
             if verbose:
                 log.info('Welcome to tenant %s', e.name)
             self.indexset_active = e
@@ -270,22 +287,24 @@ class Byneuron:
     def _data_indexsets(self, indexsets):
         """
         used for event and entity queries
-        :param indexsets: (list of) entities or indexsetKeys,
+        :param indexsets: (list of) entities or indexsetKeys, indexsetNames,
         special cases: None, use active indexset; True: use all indexsets
         :return:
         """
-        indexsetlist = indexsets if isinstance(indexsets, list) else [indexsets]
-        indexsets = set()
-        for i in indexsetlist:
+        indexset_list = indexsets if isinstance(indexsets, list) else [indexsets]
+        indexset_keys = set()
+        for i in indexset_list:
             if i is True:
-                indexsets.update({i.indexset for i in self.indexsets})  # all indexsets
+                indexset_keys.update({i.indexset for i in self.indexsets})  # all indexsets
             if isinstance(i, Entity):
-                indexsets.update({i.indexset})
+                indexset_keys.update({i.indexset})
             elif isinstance(i, str):
-                indexsets.update({i})
+                e = self.get_indexset(i)
+                if e:
+                    indexset_keys.update({e.key})
             elif i is None:
-                indexsets.update({self.indexset_active.key})
-        return list(indexsets)
+                indexset_keys.update({self.indexset_active.key})
+        return list(indexset_keys)
     
     def _data_sort(self, order):
         return 'DESC' if order == 'DESC' else 'ASC'
@@ -387,7 +406,7 @@ class Byneuron:
         return self.url(url=url, data=data)
 
     ### endpoint handlers ###
-    def get_numberevents(self, item, size, date=None, order='DESC'):
+    def get_numberevents(self, item, size=None, date=None, order='DESC'):
         """
         last 100 (size) events for item in {datetime:value} format
         starts at current timestamp (or date) and returns in time (or goes in future with order 'ASC')
@@ -420,11 +439,23 @@ class Byneuron:
                 "minInclusive": True
             }
         indexsets = indexset_keys
-        data = self._numberevents(indexsets=indexsets, filters=[f_item, f_time], size=size, order='DESC')
+        data = self._numberevents(indexsets=indexsets, filters=[f_item, f_time], size=size, order=order)
         if data:
             for i in data:
                 yield Numberevent(datamodel=i)
 
+    def get_numberevents_around(self, item, date=None, size=None):
+        """eg retrieve 100 around a certain timestamp """
+        if isinstance(size, int):
+            size = size//2
+        for event in chain(
+            self.get_numberevents(item=item, size=size, date=date, order='ASC'),
+            self.get_numberevents(item=item, size=size, date=date, order='DESC')
+        ):
+            yield event
+
+
+
     def get_numberevents_dict(self, item, size):
         return {e.datetime: e.value for e in self.get_numberevents(item=item, size=size)}
 
@@ -480,7 +511,7 @@ class Byneuron:
             else:
                 break
 
-    def get_numberevent_key(self, key, indexset=None):
+    def get_numberevent_keys(self, keys, indexset=None):
         """
         retrieve a single event using the eventid-filter
         :param key:
@@ -489,18 +520,30 @@ class Byneuron:
         """
         f = {
             "type": "eventid",
-            "values": [key],
+            "values": keys,
             "invertFilter": False,
             "operator": "or"
         }
-        data = self._numberevents(indexsets=indexset, filters=[f], size=1)
-        if data:
-            for i in data:
-                return Numberevent(datamodel=i)
-        data = self._numberevents(indexsets=indexset, filters=[f], size=1, deleted=True)
-        if data:
-            for i in data:
-                return Numberevent(datamodel=i)
+        data = self._numberevents(indexsets=indexset, filters=[f], size=len(keys), deleted=False)
+        if isinstance(data, list):
+            if len(data) != len(keys):
+                data_deleted = self._numberevents(indexsets=indexset, filters=[f], size=len(keys), deleted=True)
+                if isinstance(data_deleted, list):
+                    data.extend(data_deleted)
+            r = [Numberevent(datamodel=i) for i in data]
+            return r
+
+    def get_numberevent_key(self, key, indexset=None):
+        """
+        retrieve a single event using the eventid-filter
+        :param key:
+        :param indexset:
+        :return:
+        """
+        r = self.get_numberevent_keys([key], indexset=indexset)
+        if isinstance(r, list) and len(r):
+            return r[0]
+
 
     def get_aggregation_dates(self, item, aggrs, interval, from_date, to_date, push=None, zero=None):
         """
@@ -609,7 +652,7 @@ class Byneuron:
                 result.update({k: {}})
                 for entityKey in variables.get(f"?{k}", []):
                     e = nodes.get(entityKey, {})
-                    result[k].update({entityKey: Entity(datamodel=e)})
+                    result[k].update({entityKey: Entity(datamodel=e, indexsets=self.indexsets)})
         return result if result_with_keys else result.get(keys)
 
     def get_entities(self, entitytype, indexset=None, **kwargs):
@@ -621,6 +664,20 @@ class Byneuron:
             True, search all indexSets;
             None, search selected indexSet;
             string or [str], search indexSets by attribute-name
+        :param kwargs:
+            k:
+                'key' -> entity's key
+                'name', 'externalId', 'publicId', 'codeType' - > entity's attributes
+                'implementation' -> link:isImplementationOf Definition; to be paired with the externalId
+                'member' -> link:isMemberOf Gateway
+                'source', 'target' -> link:hasTarget, link:hasSource
+                'role', 'type' -> case deviceRole, itemType, to be paired with string (e.g. DEFAULT for deviceRole#DEFAULT)
+            v:
+                True -> any
+                None -> none
+                [] -> []
+                'regex:?' - regex:'?'
+
         """
 
         def from_kwarg(kwarg_key):
@@ -640,17 +697,60 @@ class Byneuron:
             f"{entitytype}:?e link:{'isAssignedTo' if entitytype == 'Gateway' else 'isDefinedIn'} ?indexSet."
         ]
         filter_list = []
+        # fix format used in Entity to that of datamodel
+        rename_kwarg = {'external_id':'externalId', 'public_id':'publicId', 'codetype':'codeType'}
+        for k, v in kwargs.copy().items():
+            if k in rename_kwarg:
+                kwargs.update({rename_kwarg[k]: v})
+
         for k in ['key']:
             if k in kwargs:
                 filter_list.append(f'entity:{k} {from_kwarg(k)}')
         for k in ['name', 'externalId', 'publicId', 'codeType']:
             if k in kwargs:
                 filter_list.append(f'attribute:{k} {from_kwarg(k)}')
+        for k in ['type', 'role']:
+            if k in kwargs:
+                a = f'{entitytype.lower()}{k[0].upper()}{k[1:].lower()}'
+                filter_list.append(f'attribute:{a} "{a}#{kwargs.get(k)}"')
+        for k in ['implementation']:
+            if k in kwargs:
+                v = kwargs.get(k)
+                if isinstance(v, Entity):
+                    query.append(f'{v.entity_type}:?d link:isDefinedIn ?indexSet; entity:key {v.key}.')
+                elif isinstance(v, str):
+                    query.append(f'{entitytype}Definition:?d link:isDefinedIn ?indexSet; attribute:externalId {from_kwarg(k)}.')
+                else:
+                    continue
+                filter_list.append(f'link:isImplementationOf ?d')
+        for k in ['member']:
+            if k in kwargs:
+                v = kwargs.get(k)
+                if isinstance(v, Entity):
+                    query.append(f'{entitytype}:?g link:isAssignedTo IndexSet:?is; entity:key {v.key}.')
+                elif isinstance(v, str):
+                    query.append(f'Gateway:?g link:isAssignedTo IndexSet:?is; attribute:externalId {from_kwarg(k)}.')
+                else:
+                    continue
+                filter_list.append(f'link:isMemberOf ?g')
+        for k in ['source', 'target']:
+            if k in kwargs:
+                v = kwargs.get(k)
+                if not isinstance(v, Entity):
+                    continue
+                query.append(f'{v.entity_type}:?r{k} link:isDefinedIn IndexSet:?is; entity:key {v.key}.')
+                relation_type = f'{v.entity_type}{entitytype}Relation' if k == 'source' else \
+                    f'{entitytype}{v.entity_type}Relation'
+                link1 = f'link:has{k.capitalize()} {v.entity_type}:?r{k}'
+                link2 = f'link:has{"Target" if k == "source" else "Source"} {entitytype}:?related'
+                query.append(f'{relation_type}:?r link:isDefinedIn IndexSet:?is; {link1}; {link2}.')
+                query.append(f'{entitytype}:?e intersection ?related.')
         if filter_list:
             query.append(f"{entitytype}:?e {'; '.join(filter_list)}.")
+
         offset, limit = 0, 20
         while True:
-            query.append(f"{entitytype}:?e limit {limit}; offset {offset}.")
+            query.append(f"{entitytype}:?e sort+ entity:key; limit {limit}; offset {offset}.")
             entities = self.query(query, 'e')
             for e in entities.values():
                 if isinstance(e, Entity):
@@ -660,6 +760,8 @@ class Byneuron:
             query.pop()  # remove limit
             offset += limit
 
+    # DeviceDeviceRelation:?result link:isDefinedIn #IndexSet:?index. DeviceDeviceRelation:?result link:hasTarget Device:?deviceTarget. DeviceDeviceRelation:?result link:hasSource Device:?deviceSource. DeviceDeviceRelation:?result sort+ entity:key; limit 15; offset 15.
+
     def get_entity(self, entitytype, **kwargs):
         for i in self.get_entities(entitytype, **kwargs):
             return i
@@ -683,7 +785,7 @@ class Byneuron:
                 nodes = v.get('nodes', {})
                 if k == 'entityDataModel':
                     for i in nodes.values():
-                        yield Entity(datamodel=i)
+                        yield Entity(datamodel=i, indexsets=self.indexsets)
                 elif k == 'numberEventDataModel':
                     for i in events:
                         yield Numberevent(datamodel=i)
@@ -739,7 +841,6 @@ class Byneuron:
         return [i for i in self.datamodel(datamodels)]
 
     def create_item(self, **kwargs):
-
         # to filter the keywords we can allow, and to set required fields
         item_config = {
             'description': kwargs.get('description'),
@@ -757,6 +858,12 @@ class Byneuron:
             'type': kwargs.get('type', 'DEFAULT')
         }
         e = Entity().create('Item', self.indexset_active, **{k: v for k, v in item_config.items() if v is not None})
+        if kwargs.get('definition'):
+            v = kwargs.get('definition')
+            if isinstance(v, str):
+                e = self.set_definition(e, definition=None, externalId=v)
+            elif isinstance(v, Entity):
+                e = self.set_definition(e, definition=v)
         return self.create_entity(e)
 
     def create_device(self, **kwargs):
@@ -773,26 +880,35 @@ class Byneuron:
             'roles': kwargs.get('roles', ['DEFAULT']),
             'type': kwargs.get('type', 'DEFAULT')
         }
-
         e = Entity().create('Device', self.indexset_active, **{k: v for k, v in device_config.items() if v is not None})
         if kwargs.get('definition'):
-            # add Definition using Exid, if not present, duplicate from different indexSet
-            local = self.get_entity('DeviceDefinition', externalId=kwargs.get('definition'))
-            if local is None:
-
-                remote = self.get_entity('DeviceDefinition', indexSet=True, externalId=kwargs.get('definition'))
-                if isinstance(remote, Entity):
-                    log.info('duplicate definition %s', kwargs.get('definition'))
-                    local = self.duplicate(remote, self.indexset_active)
-            if local:
-                e.set_link('isImplementationOf', local)
+            v = kwargs.get('definition')
+            if isinstance(v, str):
+                e = self.set_definition(e, definition=None, externalId=v)
+            elif isinstance(v, Entity):
+                e = self.set_definition(e, definition=v)
         # todo add profile, assetModel, status
         return self.create_entity(e)
 
+    def create_gateway(self, **kwargs):
+        device_config = {
+            'description': kwargs.get('description'),
+            'name': kwargs.get('name'),
+            'externalId': kwargs.get('externalId'),
+            'active': True,
+            'valid': True,
+            'data': kwargs.get('data'),  # optional
+            'metaKeywords': kwargs.get('metaKeywords'),  # optional
+            'metaBooleans': kwargs.get('metaBooleans'),  # optional
+            'roles': kwargs.get('roles', []),
+            'type': kwargs.get('type', 'DEFAULT')
+        }
+        e = Entity().create('Gateway', self.indexset_active, **{k: v for k, v in device_config.items() if v is not None})
+
     def create_code(self, codetype, code):
         e = Entity().create('IndexSetCode', self.indexset_active, codeType=codetype, code=code,
                             key=f'{codetype}#{code}')
-        self.create_entity(e)
+        return self.create_entity(e)
 
     def create_codetype(self, codetype, parent_codetype=None):
         e = Entity().create('IndexSetCodeType', self.indexset_active, key=codetype)
@@ -800,7 +916,35 @@ class Byneuron:
             pct = self.get_entity('IndexSetCodeType', key=parent_codetype)
             if isinstance(pct, Entity):
                 e.set_link('hasParent', pct)
-        self.create_entity(e)
+        return self.create_entity(e)
+
+    def create_relation(self, target, source):
+        if not isinstance(target, Entity):
+            return
+        if not isinstance(source, Entity):
+            return
+        if not source.roles:
+            source.set_attribute('roles', ['DEFAULT'])
+            source = self.update_entity(source)
+            if not source:
+                return
+        if not target.roles:
+            target.set_attribute('roles', ['DEFAULT'])
+            target = self.update_entity(target)
+            if not source:
+                return
+        relation_type = f'{source.entity_type}{target.entity_type}Relation'
+        source_role = source.roles[0]
+        source_role_label = f'{source.entity_type.lower()}Role'
+        source_role_full = f'{source_role_label}#{source_role}'
+        target_role = target.roles[0]
+        target_role_label = f'{target.entity_type.lower()}Role'
+        target_role_full = f'{target_role_label}#{target_role}'
+        force_key = '#'.join([source.key, target.key, source_role_full, target_role_full])
+        e = Entity().create(relation_type, self.indexset_active, key=force_key, fromDate=None, toDate=None)
+        e.set_link('hasSource', source)
+        e.set_link('hasTarget', target)
+        return self.create_entity(e)
 
     def set_features(self, parent, children):
         """
@@ -812,7 +956,24 @@ class Byneuron:
             parent.set_link('hasFeature', child)
         return self.update_entity(parent)
 
+    def remove_features(self, parent, children, delete_childeren=False):
+        """
+        :param parent: Entity of type Device, Site to remove hasFeature-relation from
+        :param children: [] of Entity of type Item to be removed hasFeature-relation
+        :return: parent after update
+        """
+        for child in children:
+            parent.del_link(child)
+        parent = self.update_entity(parent)
+        if delete_childeren:
+            for child in children:
+                self.delete_entity(child)
+        return parent
+
     def duplicate(self, entity, indexset):
+        return self.copy_to_indexset(entity, indexset)
+
+    def copy_to_indexset(self, entity, indexset):
         """
         duplicate an entity to a different indexSet
         :param entity: Entity to copy
@@ -831,6 +992,30 @@ class Byneuron:
         log.info('new entity %s from copy %s', e, entity)
         return self.create_entity(e)
 
+    def copy_entity(self, entitytype, indexset=None, **kwargs):
+        local_indexset = self.get_indexset(indexset)
+        remote_entity = self.get_entity(entitytype, indexSet=True, externalId=kwargs.get('definition'))
+        if isinstance(remote_entity, Entity):
+            log.info('duplicate %s to %s', remote_entity, local_indexset)
+            local = self.copy_to_indexset(remote_entity, local_indexset)
+            return local
+
+    def set_definition(self, entity, definition, **kwargs):
+        if not isinstance(entity, Entity):
+            return
+        definition_type = f'{entity.entity_type}Definition'
+        definition_local = None
+        if definition is None:
+            definition = self.get_entity(definition_type, **kwargs)
+        if isinstance(definition, Entity) and definition.entity_type == definition_type:
+            if definition.indexset == self.indexset_active:
+                definition_local = definition
+            else:
+                definition_local = self.copy_to_indexset(definition, self.indexset_active)
+        if definition_local:
+            entity = entity.set_link('isImplementationOf', definition_local)
+        return entity
+
     def set_isc(self, codetype, code):
         """
         routine to make sure codetypes#code are available in entity creation
@@ -866,11 +1051,12 @@ class Byneuron:
 
 
 class Entity:
-    def __init__(self, datamodel=None):
-        self._datamodel = {} if not isinstance(datamodel, dict) else datamodel
-
+    def __init__(self, datamodel=None, indexsets=None):
+        self._datamodel = datamodel if isinstance(datamodel, dict) else {}
+        self._indexsets = indexsets if isinstance(indexsets, list) else []
     def __repr__(self):
-        return f'{self.entity_type} {self.name}'
+        r = [self.entity_type, self.indexset_name, self.public_id, self.name]
+        return ' '.join([i for i in r if i != None])
 
     def wrap(self):
         return self._datamodel
@@ -1005,6 +1191,7 @@ class Entity:
         roles = self._attr(f'{self.entity_type.lower()}Role')
         if isinstance(roles, list):
             return [self.tail(i) for i in roles]
+        return []
 
     @property
     def roles(self):
@@ -1024,6 +1211,14 @@ class Entity:
             return self._link('isAssignedTo')
         return self._link('isDefinedIn')
 
+    @property
+    def indexset_name(self):
+        if self._indexsets:
+            key = self.indexset
+            for e in self._indexsets:
+                if e.key == key:
+                    return e.name
+
     @property
     def profile(self):
         return self._link('hasFeatureProfile')
@@ -1032,6 +1227,10 @@ class Entity:
     def gateway(self):
         return self._link('isMemberOfGateway')
 
+    @property
+    def definition(self):
+        return self._link('isImplementationOf')
+
     @property
     def items(self):
         return [i for i in self._links('hasFeatureItem')]
@@ -1061,7 +1260,11 @@ class Entity:
         }
         self.set_context('create')
         linktype = 'isAssignedTo' if self.entity_type == 'Gateway' else 'isDefinedIn'
-        self.set_link(linktype, indexset)
+        if isinstance(indexset, Entity) and indexset.entity_type == 'IndexSet':
+            self.set_link(linktype, indexset)
+            if not indexset in self._indexsets:
+                self._indexsets.append(indexset)
+
         for k, v in kwargs.items():
             if v is None:
                 continue
@@ -1071,15 +1274,54 @@ class Entity:
                 self.set_attribute(k, v)
         return self
 
+
+
+    def update(self, **kwargs):
+        for k, v in kwargs.items():
+            if v is None:
+                continue
+            elif k in ['description']:
+                self._datamodel.update({k: v})
+            else:
+                self.set_attribute(k, v)
+
     def set_context(self, action):
         context = self._datamodel.get('entityContext', {})
         context.update({"action": action})
         self._datamodel.update({"entityContext": context})
 
+    def _index_links(self, e):
+        """
+        find the position of entity in  _datamodel.links
+        :param e:
+        :return:
+        """
+
+        for position, _link in enumerate(self._datamodel.get('links', [])):
+            if _link.get('entityKey') == find_key:
+                return position
+
+    def del_link(self, entity):
+        """
+        remove link with this entity; also works with entity_key
+        """
+        self.set_link(linktype=None, entity=entity)
+
     def set_link(self, linktype, entity):
+        """
+        create/update/delete link with this entity
+        :param linktype:
+            e.g. hasFeature;
+            None -> remove link
+        :param entity: Entity;
+            in case of remove link this can also be an entity.key
+        :return:
+        """
+
         def _index():
+            find_key = entity.key if isinstance(entity, Entity) else entity
             for position, _link in enumerate(self._datamodel['links']):
-                if _link.get('entityKey') == entity.key:
+                if _link.get('entityKey') == find_key:
                     return position
 
         if 'links' not in self._datamodel:
@@ -1160,6 +1402,20 @@ class Entity:
                 name = f'{self.entity_type.lower()}Type'
                 type_value = f'{name}#{value}'
                 attr = _attribute(name, type_value, 'IndexSetCode', 'keywordValue')
+            elif name in ['fromDate', 'toDate']:
+                isodate = None
+                if value is None:
+                    isodate = datetime.date.today().isoformat() if name == 'fromdate' else "9999-12-31"
+                elif isinstance(value, str):
+                    try:
+                        datetime.date.isodate(value)
+                        isodate = value
+                    except ValueError:
+                        return
+                elif isinstance(value, datetime.date):
+                    isodate = value.isoformat()
+                if isodate:
+                    attr = _attribute(name, isodate, 'Date', 'dateValue')
             else:
                 return
             i = _index(name)
@@ -1190,7 +1446,7 @@ class Numberevent:
         self._datamodel = datamodel if isinstance(datamodel, dict) else {}
 
     def __repr__(self):
-        return f'Numberevent {self.datetime} {self.value}{", deleted" if self.deleted else ""}'
+        return f'Numberevent {self.datetime} {self.value}{", "+str(self.tags) if self.tags else ""}{", deleted" if self.deleted else ""}'
 
     def wrap(self):
         return self._datamodel
@@ -1224,15 +1480,24 @@ class Numberevent:
         }
         self._datamodel = datamodel
 
+    def set_value(self, value):
+        if value != value:
+            #check for nan
+            pass
+        elif isinstance(value, (float, int, bool)):
+            self._datamodel.update({"numberValue": float(value)})
+
     def set_context(self, action):
         context = self._datamodel.get('eventContext', {})
         context.update({"action": action})
         self._datamodel.update({"eventContext":context})
+
     def set_tag(self, tag):
         if 'tags' not in self._datamodel:
             self._datamodel.update({'tags': [tag]})
         elif tag not in self.tags:
             self._datamodel['tags'].append(tag)
+        return self
 
     @staticmethod
     def to_datetime(timestamp):
@@ -1319,6 +1584,30 @@ def example03():
             e = r.delete_event(e)
             log.info('deleted example event %s', e)
 
+def example04():
+    logging.basicConfig(level=logging.INFO)
+    log.info('get event')
+    # crete item 'test' in indexSet 0
+    with Byneuron() as r:
+        events = r.get_numberevent_keys(
+            keys=[
+                'b47dd72c-a2a3-43fe-a7c9-bda6634da6f5',
+                'f036989b-9a28-4c94-8ac4-4a1234d2cf74',
+                '7aa556e0-c1c1-4de0-9a04-a223c4eb7aac',
+                'cc6bc7d3-048d-4ed0-a199-3b884f34236d'
+            ],
+            indexset=True)
+        log.info('found event %s', events)
+        return
+        """
+        events = [e.set_tag('desync') for e in events]
+        print('edit', events)
+        events1 = r.update(events[:1], delete=None)
+        print('updated', events1)
+        events2 = r.update(events[1:2], delete=True)
+        print('deleted', events2)
+        """
+
 
 if __name__ == '__main__':
     example01()
